/* GStreamer
 *
 * Copyright (C) 2001-2002 Ronald Bultje <rbultje@ronald.bitfreak.net>
 *               2006 Edgard Lima <edgard.lima@indt.org.br>
 *
 * gstv4l2src.c: Video4Linux2 source element
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-v4l2src
 *
 * <refsect2>
 * v4l2src can be used to capture video from v4l2 devices, like webcams and tv cards.
 * <title>Example launch lines</title>
 * <para>
 * <programlisting>
 * gst-launch v4l2src ! xvimagesink
 * </programlisting>
 * This pipeline shows the video captured from /dev/video0 tv card and for
 * webcams.
 * </para>
 * <para>
 * <programlisting>
 * gst-launch-0.10 v4l2src ! jpegdec ! xvimagesink
 * </programlisting>
 * This pipeline shows the video captured from a webcam that delivers jpeg
 * images.
 * </para>
 * </refsect2>
 */
#define _GNU_SOURCE
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <pthread.h>
#include "v4l2src_calls.h"
#include <sys/ioctl.h>

#define GST_TSDEBUG GST_LOG

static const GstElementDetails gst_v4l2src_details =
GST_ELEMENT_DETAILS("Video (video4linux2/raw) Source",
		    "Source/Video",
		    "Reads raw frames from a video4linux2 (BT8x8) device",
		    "Ronald Bultje <rbultje@ronald.bitfreak.net>,"
		    " Edgard Lima <edgard.lima@indt.org.br>");

GST_DEBUG_CATEGORY(camsrc_debug);

enum {
	PROP_0,
	V4L2_STD_OBJECT_PROPS,
	PROP_TAKE_PICTURE,
	PROP_SET_PRV_DURATION,
	PROP_RESUME_PREVIEW,
	PROP_FLASH,
};

#define GST_TIME_AS_SECONDS(time)  ((time) / GST_SECOND)
#undef GST_TIME_AS_MSECONDS
#define GST_TIME_AS_MSECONDS(time)  ((time) / GST_MSECOND)

#ifdef ENABLE_FRAME_STORE

#define ENABLE_FRAME_HISTORY_FILE_T "/media/internal/DCIM/%s_frame-%03d-%03d.yuv"
static frame_store_seq_t *fstore_init(char *desc)
{
	frame_store_seq_t *fs = malloc(sizeof(frame_store_seq_t));
	g_assert(fs);
	fs->frame_num = fs->session_num = 0;
	fs->desc = desc;
	fs->filename = malloc(strlen(ENABLE_FRAME_HISTORY_FILE_T) + strlen(desc) + 5);
	g_assert(fs->filename);
	fs->active = FALSE;
	return fs;
}

static void fstore_uninit(frame_store_seq_t *fs)
{
	g_assert(fs && fs->filename);
	free(fs->filename);
	free(fs);
}

static void fstore_save_buf(frame_store_seq_t *fs, GstBuffer *buf, int is_still)
{
	FILE *fp;

	/* only record frames when active */
	if (!fs->active)
		return;

	g_assert(fs->filename);
	sprintf(fs->filename, ENABLE_FRAME_HISTORY_FILE_T, fs->desc,
			fs->session_num, fs->frame_num++);
	GST_INFO("saving yuv snapshot into %s", fs->filename);

	if (!(fp = fopen(fs->filename, "w"))) {
		GST_WARNING("cannot open file %s for write", fs->filename);
		goto dump_yuv_image_exit;
	}
	void *data;
	if (is_still) {
		ImageBuffer *ib = (ImageBuffer *)GST_BUFFER_DATA(buf);
		data = ib->buffer;
	}
	else
		data = GST_BUFFER_DATA(buf);
	if (fwrite(data, GST_BUFFER_SIZE(buf), 1, fp) != 1) {
		GST_WARNING("failed writing to %s", fs->filename);
		goto dump_yuv_image_exit;
	}
dump_yuv_image_exit:
	if (fp)
		fclose(fp);
}

static void fstore_session_begin(frame_store_seq_t *fs)
{
	if (!fs->active) {
		fs->active = TRUE;
		GST_INFO("starting session %d", fs->session_num);
	}
}

static void fstore_session_end(frame_store_seq_t *fs)
{
	if (fs->active) {
		GST_INFO("wrapping up session %d", fs->session_num);
		fs->session_num++;
		fs->frame_num = 0;
		fs->active = FALSE;
	}
}

#endif /* ENABLE_FRAME_STORE */

size_t fd_read(int fd, void *buf, size_t n)
{
	size_t pos = 0;
	ssize_t res;

	while (n > pos) {
		res = read(fd, buf + pos, n - pos);
		switch ((int)res) {
		case -1:
			if (errno == EINTR || errno == EAGAIN)
				continue;
			return 0;
		case 0:
			errno = EPIPE;
			return pos;
		default:
			pos += (size_t) res;
		}
	}
	return pos;
}

size_t fd_write(int fd, void *buf, size_t n)
{
	size_t pos = 0;
	ssize_t res;
	while (n > pos) {
		res = write(fd, buf + pos, n - pos);
		switch ((int)res) {
		case -1:
			if (errno == EINTR || errno == EAGAIN)
				continue;
			return 0;
		case 0:
			errno = EPIPE;
			return pos;
		default:
			pos += (size_t) res;
		}
	}
	return pos;
}

static gboolean
gst_v4l2src_iface_supported(GstImplementsInterface * iface, GType iface_type)
{
	GstV4l2Object *v4l2object = GST_V4L2SRC(iface)->v4l2object;
	return v4l2object->video_fd == -1 ? FALSE : TRUE;
}

static void gst_v4l2src_interface_init(GstImplementsInterfaceClass * klass)
{
	/*
	 * default virtual functions
	 */
	klass->supported = gst_v4l2src_iface_supported;
}

void gst_v4l2src_init_interfaces(GType type)
{
	static const GInterfaceInfo v4l2iface_info = {
		(GInterfaceInitFunc) gst_v4l2src_interface_init,
		NULL,
		NULL,
	};
	
	g_type_add_interface_static(type,
				    GST_TYPE_IMPLEMENTS_INTERFACE,
				    &v4l2iface_info);
}

GST_BOILERPLATE_FULL(GstV4l2Src, gst_v4l2src, GstPushSrc, GST_TYPE_PUSH_SRC,
		     gst_v4l2src_init_interfaces);

static void gst_v4l2src_dispose(GObject * object);
static void gst_v4l2src_finalize(GstV4l2Src * v4l2src);

/* basesrc methods */
static gboolean gst_v4l2src_start(GstBaseSrc * src);
static gboolean gst_v4l2src_stop(GstBaseSrc * src);
static GstFlowReturn gst_v4l2src_create(GstPushSrc * src, GstBuffer ** out);

static void gst_v4l2src_set_property(GObject * object, guint prop_id,
				     const GValue * value, GParamSpec * pspec);
static void gst_v4l2src_get_property(GObject * object, guint prop_id,
				     GValue * value, GParamSpec * pspec);

// singletons
static GstCaps *video_caps;
static GstCaps *still_caps;

static void gst_v4l2src_base_init(gpointer g_class)
{
	GstElementClass *gstelement_class = GST_ELEMENT_CLASS(g_class);
	GstV4l2SrcClass *gstv4l2src_class = GST_V4L2SRC_CLASS(g_class);

	gstv4l2src_class->v4l2_class_devices = NULL;

	GST_DEBUG_CATEGORY_INIT(camsrc_debug, "camsrc", 0, "camera source element");

	gst_element_class_set_details(gstelement_class, &gst_v4l2src_details);

	// video pad
	video_caps = gst_caps_new_simple ("video/x-raw-yuv",
			"format", GST_TYPE_FOURCC, GST_MAKE_FOURCC('U', 'Y', 'V', 'Y'),
			"framerate", GST_TYPE_FRACTION, 30, 1,
			"width", G_TYPE_INT, DEFAULT_VIDEO_WIDTH,
			"height", G_TYPE_INT, DEFAULT_VIDEO_HEIGHT,
			"angle", G_TYPE_INT, 270, NULL);

	gst_element_class_add_pad_template( gstelement_class,
			gst_pad_template_new("src", GST_PAD_SRC,
			GST_PAD_ALWAYS, video_caps));

	// still pad
	still_caps = gst_caps_new_simple ("video/x-raw-yuv",
		"format", GST_TYPE_FOURCC, GST_MAKE_FOURCC('U', 'Y', 'V', 'Y'),
		"width", G_TYPE_INT, DEFAULT_STILL_CAPTURE_WIDTH,
		"height", G_TYPE_INT, DEFAULT_STILL_CAPTURE_HEIGHT,
		"angle", G_TYPE_INT, 270, NULL);

	gst_element_class_add_pad_template(gstelement_class,
			gst_pad_template_new("src_still",
			GST_PAD_SRC, GST_PAD_ALWAYS, still_caps));
}

static void gst_v4l2src_class_init(GstV4l2SrcClass * klass)
{
	GObjectClass *gobject_class;
	GstBaseSrcClass *basesrc_class;
	GstPushSrcClass *pushsrc_class;

	gobject_class = G_OBJECT_CLASS(klass);
	basesrc_class = GST_BASE_SRC_CLASS(klass);
	pushsrc_class = GST_PUSH_SRC_CLASS(klass);

	gobject_class->dispose = gst_v4l2src_dispose;
	gobject_class->finalize = (GObjectFinalizeFunc) gst_v4l2src_finalize;
	gobject_class->set_property = gst_v4l2src_set_property;
	gobject_class->get_property = gst_v4l2src_get_property;

	gst_v4l2_object_install_properties_helper(gobject_class);

	g_object_class_install_property(gobject_class, PROP_TAKE_PICTURE,
					g_param_spec_boolean("take-picture",
							     "Take Picture",
							     "Take a still picture",
							     FALSE,
							     G_PARAM_WRITABLE));

	g_object_class_install_property(gobject_class, PROP_RESUME_PREVIEW,
					g_param_spec_boolean("resume-preview",
							     "Resume Preview",
							     "Resume Preview after still capture",
							     FALSE,
							     G_PARAM_WRITABLE));

	g_object_class_install_property(gobject_class, PROP_SET_PRV_DURATION,
					g_param_spec_uint("preview-duration",
							  "Preview Duration",
							  "Preview Duration of still captured frame in milli seconds",
							  0, 0xffffffff, 0,
							  G_PARAM_READWRITE));

	g_object_class_install_property(gobject_class, PROP_FLASH,
					g_param_spec_uint("flash-ctrl",
							  "flash Control",
							  "Command to control the flash",
							  0, 2, 0,
							  G_PARAM_READWRITE));

	basesrc_class->start = GST_DEBUG_FUNCPTR(gst_v4l2src_start);
	basesrc_class->stop = GST_DEBUG_FUNCPTR(gst_v4l2src_stop);

	pushsrc_class->create = GST_DEBUG_FUNCPTR(gst_v4l2src_create);
}

static void post_still_preview_unblock(GstV4l2Src *v4l2src)
{
	g_mutex_lock(v4l2src->post_still_hold_frame_lock);
	if (v4l2src->pausing_after_still) {
		g_cond_signal(v4l2src->still_hold_frame_cond);
		GST_INFO("unblocking post-capture wait");
	}
	else
		GST_DEBUG("not paused, skipping");
	g_mutex_unlock(v4l2src->post_still_hold_frame_lock);
}

static struct v4l2_fmtdesc *gst_v4l2src_get_format_from_fourcc(GstV4l2Src *v4l2src, 
		guint32 fourcc)
{
	struct v4l2_fmtdesc *fmt;
	GSList *walk;

	if (fourcc == 0)
		return NULL;

	walk = v4l2src->formats;
	while (walk) {
		fmt = (struct v4l2_fmtdesc *)walk->data;
		if (fmt->pixelformat == fourcc)
			return fmt;
		/* special case for jpeg */
		if ((fmt->pixelformat == V4L2_PIX_FMT_MJPEG && fourcc == V4L2_PIX_FMT_JPEG)
		    || (fmt->pixelformat == V4L2_PIX_FMT_JPEG
			&& fourcc == V4L2_PIX_FMT_MJPEG)) {
			return fmt;
		}
		walk = g_slist_next(walk);
	}

	return NULL;
}

/* collect data for the given caps
 * @caps: given input caps
 * @format: location for the v4l format
 * @w/@h: location for width and height
 * @fps_n/@fps_d: location for framerate
 * @size: location for expected size of the frame or 0 if unknown
 */
static gboolean gst_v4l2_get_caps_info(GstV4l2Src * v4l2src, GstCaps * caps,
		       struct v4l2_fmtdesc **format, gint * w, gint * h,
		       guint * fps_n, guint * fps_d, guint * size)
{
	GstStructure *structure;
	const GValue *framerate;
	guint32 fourcc;
	const gchar *mimetype;
	guint outsize;

	/* default unknown values */
	fourcc = 0;
	outsize = 0;

	structure = gst_caps_get_structure(caps, 0);

	if (!gst_structure_get_int(structure, "width", w))
		return FALSE;

	if (!gst_structure_get_int(structure, "height", h))
		return FALSE;

	framerate = gst_structure_get_value(structure, "framerate");
	if (!framerate)
		return FALSE;

	*fps_n = gst_value_get_fraction_numerator(framerate);
	*fps_d = gst_value_get_fraction_denominator(framerate);

	mimetype = gst_structure_get_name(structure);

	if (!strcmp(mimetype, "video/x-raw-yuv")) {
		if (!gst_structure_get_fourcc(structure, "format", &fourcc))
			return FALSE;

		switch (fourcc) {
		case GST_MAKE_FOURCC('I', '4', '2', '0'):
		case GST_MAKE_FOURCC('I', 'Y', 'U', 'V'):
			fourcc = V4L2_PIX_FMT_YUV420;
			outsize = GST_ROUND_UP_4(*w) * GST_ROUND_UP_2(*h);
			outsize +=
			    2 * ((GST_ROUND_UP_8(*w) / 2) *
				 (GST_ROUND_UP_2(*h) / 2));
			break;
		case GST_MAKE_FOURCC('Y', 'U', 'Y', '2'):
			fourcc = V4L2_PIX_FMT_YUYV;
			outsize = (GST_ROUND_UP_2(*w) * 2) * *h;
			break;
		case GST_MAKE_FOURCC('Y', '4', '1', 'P'):
			fourcc = V4L2_PIX_FMT_Y41P;
			outsize = (GST_ROUND_UP_2(*w) * 2) * *h;
			break;
		case GST_MAKE_FOURCC('U', 'Y', 'V', 'Y'):
			fourcc = V4L2_PIX_FMT_UYVY;
			outsize = (GST_ROUND_UP_2(*w) * 2) * *h;
			break;
		case GST_MAKE_FOURCC('Y', 'V', '1', '2'):
			fourcc = V4L2_PIX_FMT_YVU420;
			outsize = GST_ROUND_UP_4(*w) * GST_ROUND_UP_2(*h);
			outsize +=
			    2 * ((GST_ROUND_UP_8(*w) / 2) *
				 (GST_ROUND_UP_2(*h) / 2));
			break;
		case GST_MAKE_FOURCC('Y', '4', '1', 'B'):
			fourcc = V4L2_PIX_FMT_YUV411P;
			outsize = GST_ROUND_UP_4(*w) * *h;
			outsize += 2 * ((GST_ROUND_UP_8(*w) / 4) * *h);
			break;
		case GST_MAKE_FOURCC('Y', '4', '2', 'B'):
			fourcc = V4L2_PIX_FMT_YUV422P;
			outsize = GST_ROUND_UP_4(*w) * *h;
			outsize += 2 * ((GST_ROUND_UP_8(*w) / 2) * *h);
			break;
		}
	} else if (!strcmp(mimetype, "video/x-raw-rgb")) {
		gint depth, endianness, r_mask;

		if (!gst_structure_get_int(structure, "depth", &depth))
			return FALSE;
		if (!gst_structure_get_int(structure, "endianness", &endianness))
			return FALSE;
		if (!gst_structure_get_int(structure, "red_mask", &r_mask))
			return FALSE;

		switch (depth) {
		case 8:
			fourcc = V4L2_PIX_FMT_RGB332;
			break;
		case 15:
			fourcc = (endianness == G_LITTLE_ENDIAN) ?
			    V4L2_PIX_FMT_RGB555 : V4L2_PIX_FMT_RGB555X;
			break;
		case 16:
			fourcc = (endianness == G_LITTLE_ENDIAN) ?
			    V4L2_PIX_FMT_RGB565 : V4L2_PIX_FMT_RGB565X;
			break;
		case 24:
			fourcc =
			    (r_mask ==
			     0xFF) ? V4L2_PIX_FMT_BGR24 : V4L2_PIX_FMT_RGB24;
			break;
		case 32:
			fourcc =
			    (r_mask ==
			     0xFF) ? V4L2_PIX_FMT_BGR32 : V4L2_PIX_FMT_RGB32;
			break;
		}
	} else if (strcmp(mimetype, "video/x-dv") == 0) {
		fourcc = V4L2_PIX_FMT_DV;
	} else if (strcmp(mimetype, "image/jpeg") == 0) {
		fourcc = V4L2_PIX_FMT_JPEG;
	}

	if (fourcc == 0)
		return FALSE;

	if (!v4l2src->formats)
		gst_v4l2src_fill_format_list(v4l2src);
	*format = gst_v4l2src_get_format_from_fourcc(v4l2src, fourcc);
	*size = outsize;

	return TRUE;
}

static gboolean gst_pad_event_handler(GstPad* pad, GstEvent* event)
{
	gboolean ret = TRUE;
	GstV4l2Src *v4l2src = GST_V4L2SRC(GST_PAD_PARENT(pad));

	switch (GST_EVENT_TYPE(event)) {
		case GST_EVENT_QOS:
			break;
		case GST_EVENT_CUSTOM_UPSTREAM:
			GST_LOG("got event %s, assuming buffer is ready for Q", 
					gst_event_type_get_name(GST_EVENT_TYPE(event)));
			CAMSRC_W_NOTIFICATION(CAMSRC_N_BUFFER_READY);
			break;
		default:
			GST_INFO("not handling event type %s", 
					gst_event_type_get_name(GST_EVENT_TYPE(event)));
			ret = FALSE;
			break;
	}
	return ret;
}

#define NUM_FRAMES_SKIPPED_ON_SESSION_START 3

gboolean gst_v4l2_ready_to_pasued(GstV4l2Src *v4l2src)
{
	gint w = 0, h = 0;
	struct v4l2_fmtdesc *format = NULL;
	guint fps_n, fps_d;
	guint size;
	gboolean ret = FALSE;

	PROF_SECT_DECL_BEGIN(init_camera_session_and_preview);

	g_assert(GST_V4L2_IS_OPEN(v4l2src->v4l2object));

	/* set fmt and request the still buffers before going into preview */
	if (!gst_v4l2src_set_still_capture(v4l2src))
		goto gst_v4l2_ready_to_pasued_exit;
	if (!still_capture_buff_init(v4l2src, still_caps, v4l2src->v4l2object->still_fd))
		goto gst_v4l2_ready_to_pasued_exit;

	/* we want our own v4l2 type of fourcc codes */
	if (!gst_v4l2_get_caps_info(v4l2src, video_caps,
				&format, &w, &h, &fps_n, &fps_d, &size)) {
		GST_DEBUG_OBJECT(v4l2src, "can't get capture format from video caps %"
				 GST_PTR_FORMAT, video_caps);
		goto gst_v4l2_ready_to_pasued_exit;
	}

	g_assert(format != NULL);
    
	GST_DEBUG_OBJECT(v4l2src, "resolution: %dx%d, %d/%d fps (max), "
			 "format %s", w, h, fps_n, fps_d, format->description);

	v4l2src->fps_n = fps_n;
	v4l2src->fps_d = fps_d;
	v4l2src->width = w;
	v4l2src->height = h;
	v4l2src->pixelformat = format->pixelformat;
	v4l2src->num_discard_video_frames = NUM_FRAMES_SKIPPED_ON_SESSION_START;

	memset(v4l2src->video_buf, 0, sizeof(v4l2src->video_buf));

	if (!gst_v4l2src_set_capture(v4l2src, format->pixelformat, w, h, fps_n, fps_d)) {
		/* error already posted */
		goto gst_v4l2_ready_to_pasued_exit;
	}
	struct v4l2_requestbuffers breq;

	memset(&breq, 0, sizeof(struct v4l2_requestbuffers));
	breq.count = NUM_OVERLAY_BUFFERS;
	breq.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	breq.memory = V4L2_MEMORY_USERPTR;

	if (ioctl(v4l2src->v4l2object->video_fd, VIDIOC_REQBUFS, &breq) < 0) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				  (_("VIDIOC_REQBUFS failed")), GST_ERROR_SYSTEM);
		goto gst_v4l2_ready_to_pasued_exit;
	}
	GST_V4L2_SET_ACTIVE(v4l2src->v4l2object);

	/* now store the expected output size */
	v4l2src->frame_byte_size = size;
	v4l2src->frame_create_mode = FCM_PREVIEW_1ST_FRAME;

	/* inter-threads notifications */
	v4l2src->event_info[PE_NOTIFY].fd = v4l2src->v4l2object->propset_notify_fd[0];
	v4l2src->relinquish_still_buffer = FALSE;
	gst_pad_set_event_function(GST_BASE_SRC(v4l2src)->srcpad, gst_pad_event_handler);
	
	ret = TRUE;
gst_v4l2_ready_to_pasued_exit:
	PROF_SECT_END(init_camera_session_and_preview);
	return ret;

}

#ifdef YUVS_ENABLE

#define YUVS_MAX_QUQUE_SIZE 9
#define YUVS_FILE "/media/internal/DCIM/yuvs.bin"
#define YUVS_FRAME_SIZE (DEFAULT_VIDEO_WIDTH * DEFAULT_VIDEO_HEIGHT * 2)
#define YUVS_FRAME_SAVE_RATIO 5

static void *yuvs_bg_store_thread(void *arg)
{
	GstV4l2Src *v4l2src = arg;
	GST_DEBUG("TBD: open file\n");
	FILE *fp = fopen(YUVS_FILE, "w");

	int count = 0;
	gboolean try_save = TRUE;

	g_assert(fp);

	while (1) {
		g_assert(!sem_wait(&v4l2src->yuvs.sem));
		if (v4l2src->yuvs.shutdown)
			break;
		int val;
		g_assert(!sem_getvalue(&v4l2src->yuvs.sem, &val));
		g_assert(val >= 0 && val < YUVS_MAX_QUQUE_SIZE);

		g_mutex_lock(v4l2src->yuvs.lock);
		GstBuffer *buf = g_queue_pop_tail(v4l2src->yuvs.queue);
		g_mutex_unlock(v4l2src->yuvs.lock);

		/* save only 50% of buffers to maintain consistant frame rate */
		if (((count++ % YUVS_FRAME_SAVE_RATIO) == 0) && try_save) {
			if (fwrite(GST_BUFFER_DATA(buf), 
						YUVS_FRAME_SIZE, 1, fp) != 1) {
				GST_WARNING("can't write to %s", YUVS_FILE);
				try_save = FALSE;
			}
			else
				GST_DEBUG("buffer %p saved (#%d)", buf, count);
		}
		gst_buffer_unref(buf);
	}
	GST_DEBUG("file %s closed", YUVS_FILE);
	g_assert(!fclose(fp));
	return NULL;
}

#endif

static GstStateChangeReturn gst_v4l2src_change_state(GstElement* element,
		GstStateChange transition)
{
	GstV4l2Src *v4l2src = GST_V4L2SRC(element);
	GstStateChangeReturn ret = GST_STATE_CHANGE_FAILURE;

	GST_DEBUG ("tid[%d]: plugin state changes from %s to %s (element state: %s)",
			syscall(SYS_gettid),
			gst_element_state_get_name(GST_STATE_TRANSITION_CURRENT(transition)),
			gst_element_state_get_name(GST_STATE_TRANSITION_NEXT(transition)),
			gst_element_state_get_name (GST_STATE(GST_ELEMENT(v4l2src))));

	switch (transition) {
		case GST_STATE_CHANGE_NULL_TO_READY:
			GST_LOG("GST_STATE_CHANGE_NULL_TO_READY");
			/* open the device */
			if (!gst_v4l2_object_start(v4l2src->v4l2object))
				goto gst_v4l2src_change_state_exit;
			break;
		case GST_STATE_CHANGE_READY_TO_PAUSED:
			GST_LOG("GST_STATE_CHANGE_READY_TO_PAUSED");
#ifdef YUVS_ENABLE
			g_assert(!sem_init(&v4l2src->yuvs.sem, 0, 0));
			v4l2src->yuvs.shutdown = FALSE;
			v4l2src->yuvs.queue = g_queue_new();
			g_assert(!pthread_create(&v4l2src->yuvs.thread_id, 
					NULL, yuvs_bg_store_thread, v4l2src));
#endif
			if (!gst_v4l2_ready_to_pasued(v4l2src)) {
				ret = GST_STATE_CHANGE_FAILURE;
				goto gst_v4l2src_change_state_exit;
			}
			break;
		case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
			GST_LOG("GST_STATE_CHANGE_PAUSED_TO_PLAYING");
			if (!gst_v4l2src_capture_start(v4l2src)) {
				ret = GST_STATE_CHANGE_FAILURE;
				goto gst_v4l2src_change_state_exit;
			}
			GST_INFO("video preview resumed");
			break;
		default:
			break;
	}

	/* "preventative" measure - if preview frame is being paused on screen,
	 * we must unblock it here so that parent->change_state() does not hang
	 */
	if (transition == GST_STATE_CHANGE_PLAYING_TO_PAUSED) {
		post_still_preview_unblock(v4l2src);
		CAMSRC_W_NOTIFICATION(CAMSRC_N_PAUSED);
	}

	GST_LOG ("before parent element state change");
	ret = GST_ELEMENT_CLASS(parent_class)->change_state(element, transition);
	if (ret != GST_STATE_CHANGE_SUCCESS && ret != GST_STATE_CHANGE_NO_PREROLL) {
		goto gst_v4l2src_change_state_exit;
	}
	GST_LOG ("parent element state change done");

	switch (transition) {
		case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
			GST_LOG("GST_STATE_CHANGE_PLAYING_TO_PAUSED");
			if (!gst_v4l2src_capture_stop(v4l2src)) {
				ret = GST_STATE_CHANGE_FAILURE;
				goto gst_v4l2src_change_state_exit;
			}
			GST_INFO("video preview paused");
			break;
		case GST_STATE_CHANGE_PAUSED_TO_READY:
			GST_LOG("GST_STATE_CHANGE_PAUSED_TO_READY");
			v4l2src->relinquish_still_buffer = TRUE;
			GST_DEBUG("about to free still capture buffer, refount=%d ",
					GST_MINI_OBJECT_REFCOUNT_VALUE(
					v4l2src->still_capture_buffer));
			gst_buffer_unref(v4l2src->still_capture_buffer);
			GST_DEBUG("after unref, refount=%d ",
					GST_MINI_OBJECT_REFCOUNT_VALUE(
					v4l2src->still_capture_buffer));
			v4l2src->still_capture_buffer = NULL;
#ifdef YUVS_ENABLE
			v4l2src->yuvs.shutdown = TRUE;
			sem_post(&v4l2src->yuvs.sem);
			void *ev;
			pthread_join(v4l2src->yuvs.thread_id, &ev);
			g_queue_free(v4l2src->yuvs.queue);
			v4l2src->yuvs.queue = NULL;
			g_assert(!sem_destroy(&v4l2src->yuvs.sem));
#endif
			break;
		case GST_STATE_CHANGE_READY_TO_NULL:
			GST_LOG("GST_STATE_CHANGE_READY_TO_NULL");
			/* close the device */
			if (!gst_v4l2_object_stop(v4l2src->v4l2object))
				goto gst_v4l2src_change_state_exit;
			break;
		default:
			break;
	}

gst_v4l2src_change_state_exit:
	GST_LOG("end (ret=%d)", ret);
	return ret;
}

PROF_EVENT_DEF_GLBL(sipc_init_to_connect);

static void gst_v4l2src_init(GstV4l2Src * v4l2src, GstV4l2SrcClass * klass)
{
	GstElementClass*   gstelement_class;

	gstelement_class = (GstElementClass *) klass;
	gstelement_class->change_state = gst_v4l2src_change_state;
	v4l2src->v4l2object = gst_v4l2_object_new(GST_ELEMENT(v4l2src),
						  gst_v4l2_get_input,
						  gst_v4l2_set_input, NULL);

	if (!v4l2src->v4l2object)
		goto gst_v4l2src_init_exit;

	if (!(v4l2src->h_sipc = sipc_create())) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE, (_("sipc create error")),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_init_err_sipc_create;
	}
	PROF_SECT_DECL_BEGIN(sipc_init);
	if (sipc_init(v4l2src->h_sipc, SIPC_MODE_CLIENT, CAMD_IPC_ADDR)) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE, (_("%s"),
					sipc_get_err_str(v4l2src->h_sipc)),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_init_err_sipc_init;
	}
	const char *sipc_err = sipc_get_err_str(v4l2src->h_sipc);
	if (sipc_err)
		GST_WARNING("sipc_init completed w/warning: %s", sipc_err);

	PROF_SECT_END(sipc_init);
	PROF_BEGIN(sipc_init_to_connect);

	v4l2src->formats = NULL;

	gst_base_src_set_format(GST_BASE_SRC(v4l2src), GST_FORMAT_TIME);
	gst_base_src_set_live(GST_BASE_SRC(v4l2src), TRUE);

	memset(&(v4l2src->still_format), 0, sizeof(struct v4l2_format));

	v4l2src->still_format.fmt.pix.width = DEFAULT_STILL_CAPTURE_WIDTH;
	v4l2src->still_format.fmt.pix.height = DEFAULT_STILL_CAPTURE_HEIGHT;

	v4l2src->flash_required = FALSE;
	v4l2src->flash_ctrl = FLASH_OFF;

	v4l2src->src_still_pad = gst_pad_new_from_template(
			gst_element_class_get_pad_template((GstElementClass *) klass,
			"src_still"), "src_still");

	GST_DEBUG_OBJECT(v4l2src, "adding src_still pad");
	gst_element_add_pad(GST_ELEMENT(v4l2src), v4l2src->src_still_pad);

	v4l2src->mutex = g_mutex_new();
	if (sem_init(&v4l2src->still_capture_sem, 0, 1))
		goto gst_v4l2src_init_err_sem_init;

	// Some more additions
	v4l2src->still_duration = 0;
	v4l2src->fps_d = 0;
	v4l2src->fps_n = 0;
	v4l2src->camd_comm_lock = g_mutex_new();
	v4l2src->notification_lock = g_mutex_new();
	v4l2src->post_still_hold_frame_lock = g_mutex_new();
	v4l2src->still_buffer_lock = g_mutex_new();
	v4l2src->still_hold_frame_cond = g_cond_new();

#ifdef STORE_PRE_STILL_FRAMES
	v4l2src->preview_fh = fstore_init("preview");
#endif	
#ifdef STORE_STILL_FRAMES
	v4l2src->still_fh = fstore_init("still");
#endif	
#ifdef YUVS_ENABLE
	v4l2src->yuvs.lock = g_mutex_new();
#endif
	v4l2src->init_complete = TRUE;
	goto gst_v4l2src_init_exit;

gst_v4l2src_init_err_sem_init:

gst_v4l2src_init_err_sipc_init:
	if (sipc_uninit(v4l2src->h_sipc)) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("sipc uninit error: %s"),
				 sipc_get_err_str(v4l2src->h_sipc)),
				GST_ERROR_SYSTEM);
		v4l2src->h_sipc = NULL;
	}

gst_v4l2src_init_err_sipc_create:
	gst_v4l2_object_destroy(v4l2src->v4l2object);

gst_v4l2src_init_exit:
	return;
}

static void gst_v4l2src_dispose(GObject * object)
{
	G_OBJECT_CLASS(parent_class)->dispose(object);
}

static void gst_v4l2src_finalize(GstV4l2Src * v4l2src)
{
	if (!v4l2src->init_complete)
		return;

#ifdef STORE_PRE_STILL_FRAMES
	fstore_uninit(v4l2src->preview_fh);
#endif
#ifdef STORE_STILL_FRAMES
	fstore_uninit(v4l2src->still_fh);
#endif
	if (v4l2src->formats) {
		gst_v4l2src_clear_format_list(v4l2src);
	}

	g_cond_free(v4l2src->still_hold_frame_cond);
	g_mutex_free(v4l2src->post_still_hold_frame_lock);
	g_mutex_free(v4l2src->still_buffer_lock );
	g_mutex_free(v4l2src->mutex);
	v4l2src->mutex = NULL;
	g_mutex_free(v4l2src->camd_comm_lock);
	v4l2src->camd_comm_lock = NULL;
	g_mutex_free(v4l2src->notification_lock);
	v4l2src->notification_lock = NULL;
#ifdef YUVS_ENABLE
	g_mutex_free(v4l2src->yuvs.lock);
	v4l2src->yuvs.lock = NULL;
#endif

	if (v4l2src->h_sipc && sipc_uninit(v4l2src->h_sipc)) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("sipc uninit error: %s"),
				 sipc_get_err_str(v4l2src->h_sipc)),
				GST_ERROR_SYSTEM);
	}
	v4l2src->event_info[PE_IPC].fd = -1;

	gst_v4l2_object_destroy(v4l2src->v4l2object);

	G_OBJECT_CLASS(parent_class)->finalize((GObject *) (v4l2src));
	v4l2src->init_complete = FALSE;
}

PROF_EVENT_DEF_GLBL(wait_for_preview_frame);
PROF_EVENT_DEF_GLBL(start_still_capture_lag);
PROF_EVENT_DEF_GLBL(still_exposure_and_readout);
PROF_EVENT_DEF_GLBL(torch_converge);
PROF_EVENT_DEF_GLBL(still_start_until_ready);

static void
gst_v4l2src_set_property(GObject * object,
			 guint prop_id, const GValue * value,
			 GParamSpec * pspec)
{
	GstV4l2Src *v4l2src = GST_V4L2SRC(object);

	if (!v4l2src->init_complete) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SYNC,
				(_("set_property of prop_id %u rcvd, "
				   "but init incomplete!"),
				 prop_id), GST_ERROR_SYSTEM);
		return;
	}

	if (!gst_v4l2_object_set_property_helper(v4l2src->v4l2object,
						 prop_id, value, pspec)) {
		g_mutex_lock(v4l2src->mutex);
		switch (prop_id) {
		case PROP_TAKE_PICTURE:

			if (v4l2src->frame_create_mode == FCM_PREVIEW &&
					!sem_trywait(&v4l2src->still_capture_sem)) {
				GST_INFO_OBJECT(v4l2src, "PROP_TAKE_PICTURE");
				PROF_BEGIN(start_still_capture_lag);
				CAMSRC_W_NOTIFICATION(CAMSRC_N_STILL_CAPTURE);
			}
			else {
				GST_ELEMENT_ERROR(v4l2src, RESOURCE, SYNC,
						(_("PROP_TAKE_PICTURE received while still "
						   "capture already in progress")),
						GST_ERROR_SYSTEM);
			}
			break;
		case PROP_SET_PRV_DURATION:
			v4l2src->still_duration = g_value_get_uint(value);
			GST_INFO("PROP_SET_PRV_DURATION=%d", v4l2src->still_duration);
			break;
		case PROP_RESUME_PREVIEW:
			GST_INFO("PROP_RESUME_PREVIEW rcvd");
			post_still_preview_unblock(v4l2src);
			break;
		case PROP_FLASH:
			v4l2src->flash_ctrl = g_value_get_uint(value);
			GST_INFO_OBJECT(v4l2src, "PROP_FLASH=%d", v4l2src->flash_ctrl);
			CAMSRC_W_NOTIFICATION(CAMSRC_N_FLASH_MODE);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id,
							  pspec);
			break;
		}
		g_mutex_unlock(v4l2src->mutex);
	}
}

static void
gst_v4l2src_get_property(GObject * object,
			 guint prop_id, GValue * value, GParamSpec * pspec)
{
	GstV4l2Src *v4l2src = GST_V4L2SRC(object);

	if (!gst_v4l2_object_get_property_helper(v4l2src->v4l2object,
						 prop_id, value, pspec)) {
		switch (prop_id) {
		case PROP_FLASH:
			g_value_set_uint(value, v4l2src->flash_ctrl);
			break;
		case PROP_SET_PRV_DURATION:
			g_value_set_uint(value, v4l2src->still_duration);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id,
							  pspec);
			break;
		}
	}
}

/* start and stop are not symmetric -- start will open the device, but not start
 * capture. it's setcaps that will start capture, which is called via basesrc's
 * negotiate method. stop will both stop capture and close the device.
 */
static gboolean gst_v4l2src_start(GstBaseSrc * src)
{
	gboolean ret = FALSE;
	GstV4l2Src *v4l2src = GST_V4L2SRC(src);
	if (!(v4l2src->h_camd_msg = camd_msg_init(v4l2src->h_sipc))) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE, (_("camd msg init error")),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_start_exit;
	}

	PROF_END(sipc_init_to_connect);
	PROF_SECT_DECL_BEGIN(connect_to_camd);
	if (sipc_client_connect_timeout(v4l2src->h_sipc, -1)) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("error connecting to camd: %s"),
				 sipc_get_err_str(v4l2src->h_sipc)),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_start_exit;
	}
	PROF_SECT_END(connect_to_camd);

	/* ipc events */
	v4l2src->event_info[PE_IPC].fd = sipc_get_connection_fd(v4l2src->h_sipc);

	v4l2src->offset = 0;

	ret = TRUE;
gst_v4l2src_start_exit:
	return ret;
}

static gboolean gst_v4l2src_stop(GstBaseSrc * src)
{
	gboolean ret = FALSE;
	PROF_SECT_DECL_BEGIN(gst_v4l2src_stop);
	GstV4l2Src *v4l2src = GST_V4L2SRC(src);

	if (v4l2src->v4l2object->buffer != NULL) {
		if (!gst_v4l2src_capture_deinit(v4l2src))
			goto gst_v4l2src_capture_stop_exit;
	}

	camd_msg_uninit(v4l2src->h_camd_msg);

	if (sipc_disconnect(v4l2src->h_sipc)) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("%s"), sipc_get_err_str(v4l2src->h_sipc)),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_capture_stop_exit;
	}
	v4l2src->event_info[PE_IPC].fd = -1;

	v4l2src->fps_d = 0;
	v4l2src->fps_n = 0;

	ret = TRUE;

gst_v4l2src_capture_stop_exit:
	PROF_SECT_END(gst_v4l2src_stop);
	return ret;
}

#if defined(TRACE_POLL_EVENTS) && defined(DEBUG)
#define EVENT_ENTRY(e) {e, #e}
#define TRACE_LINE_LEN 128

static void trace_pollfd_events(struct pollfd *p, int num_fds)
{
	/* possbile events */
	static struct {
		int event;
		char *str;
	} pevents[] = {
		EVENT_ENTRY(POLLIN),
		EVENT_ENTRY(POLLOUT),
		EVENT_ENTRY(POLLHUP),
		EVENT_ENTRY(POLLNVAL),
		EVENT_ENTRY(POLLERR),
		EVENT_ENTRY(POLLPRI),
		{0, NULL}
	};
	int i;
	char str[TRACE_LINE_LEN];

	int len = 0;
	for (i = 0; i < num_fds; i++) {
		if (!p[i].revents)
			continue;

		len += sprintf(str + len, "fd=%d: revents=(", p[i].fd);
		int j, printed = 0;
		for (j = 0; pevents[j].event; j++) {
			if (p[i].revents & pevents[j].event) {
				len += sprintf(str + len,
						printed ? ",%s" : "%s",
						pevents[j].str);
				printed = 1;
			}
		}
		g_assert(printed);
		len += sprintf(str + len, ") ");
	}
	g_assert(len + 1 < TRACE_LINE_LEN);
	GST_LOG("%s", str);
}
#endif /* DEBUG */

#define GET_EVENTS_TIMEOUT 1000

typedef enum {
	GE_EVENTS_RECEIVED = 0,
	GE_TIMEOUT,
	GE_FATAL_ERROR,
} get_event_result_t;

static get_event_result_t get_events(GstV4l2Src *v4l2src)
{
	int i, num_events = v4l2src->num_monitored_poll_events;
	struct pollfd *pem = v4l2src->pem;
	get_event_result_t ret = GE_EVENTS_RECEIVED;

	for (i =0; i < num_events; i++)
		pem[i].events = POLLIN;

	while (1) {
		for (i = 0; i < num_events; i++)
			pem[i].revents = 0;
		int res = poll(pem, num_events, GET_EVENTS_TIMEOUT);

		if (!res)
			goto timeout;
		if (res == -1) {
			if (errno == EINTR)
				continue;
			goto poll_failed;
		}
		/* success, some events received */
		break;
	}

	/* verify no errors on any of the fds */
	for (i = 0; i < num_events; i++) {
		if (pem[i].revents & POLLERR)
			goto error_event;
	}

#if defined(TRACE_POLL_EVENTS) && defined(DEBUG)
	trace_pollfd_events(pem, num_events);
#endif
	goto get_events_exit;
timeout:
	{
		GST_WARNING("poll timeout: no activity in %dms", GET_EVENTS_TIMEOUT);
		ret = GE_TIMEOUT;
		goto get_events_exit;
	}
poll_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE, (_("poll() failed")),
				GST_ERROR_SYSTEM);
		ret = GE_FATAL_ERROR;
		goto get_events_exit;
	}
error_event:
	{
		int id;
		for (id = 0; id < PE_LAST; id++) {
			if (v4l2src->event_info[id].fd == pem[i].fd)
				break;
		}

		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("error event when polling on fd=%d, id=%d"
				" num_queued_buffers=%d"), 
				pem[i].fd, id, v4l2src->num_queued_buffers),
				GST_ERROR_SYSTEM);
		ret = GE_FATAL_ERROR;
		goto get_events_exit;
	}
get_events_exit:
	return ret;
}

#define EVENT_MONITOR_SECT_START \
{ \
	int i = 0; \
	v4l2src->num_monitored_poll_events = 0; \
	struct pollfd *pem = v4l2src->pem;

#define EVENT_MONITOR_SECT_END \
	v4l2src->num_monitored_poll_events = i; \
}

#define EVENT_ADD_MONITOR(id) \
	do { \
		pem[i].fd = v4l2src->event_info[id].fd; \
		pem[i].revents = 0; \
		v4l2src->event_info[id].poll_idx = i; \
		i++; \
	} while (0)

#define EVENTS(id) v4l2src->pem[v4l2src->event_info[id].poll_idx].revents

static GstFlowReturn setup_events_preview(GstV4l2Src *v4l2src)
{
	int ret = GST_FLOW_OK;

	v4l2src->event_info[PE_V4L2].fd = v4l2src->v4l2object->video_fd;
	EVENT_MONITOR_SECT_START;

	if (v4l2src->num_queued_buffers) {
#ifdef SHOW_FRAMES_QDQ
		GST_LOG("will select on v4l2 events (%d buffers queued)",
				v4l2src->num_queued_buffers);
#endif
		EVENT_ADD_MONITOR(PE_V4L2);
	}
	else
		GST_WARNING("no buffers queued, large frame-timestamp gap possible");

	EVENT_ADD_MONITOR(PE_NOTIFY);
	EVENT_ADD_MONITOR(PE_IPC);
	EVENT_MONITOR_SECT_END;
	return ret;
}

static void setup_events_still(GstV4l2Src *v4l2src)
{
	v4l2src->event_info[PE_V4L2].fd = v4l2src->v4l2object->still_fd;
	EVENT_MONITOR_SECT_START;
	EVENT_ADD_MONITOR(PE_V4L2);
	EVENT_ADD_MONITOR(PE_IPC);
	EVENT_MONITOR_SECT_END;
}

static GstFlowReturn gst_v4l2src_still_capture_start(GstV4l2Src *v4l2src)
{
	GstFlowReturn ret = GST_FLOW_ERROR;

	PROF_BEGIN(still_start_until_ready);
				
	if (v4l2src->still_duration) {
		/* post-still capture delay requested,
		 * mark time when still capture begins
		 */
		GST_OBJECT_LOCK(v4l2src);
		v4l2src->still_capture_start_ts =
			gst_clock_get_time(GST_ELEMENT_CLOCK(v4l2src));
		GST_OBJECT_UNLOCK(v4l2src);
	}

	if (v4l2src->current_flash_ctrl == FLASH_ON ||
			(v4l2src->current_flash_ctrl == FLASH_AUTO &&
			 v4l2src->flash_required)) {
		if (CAMD_NOTIFICATION_SEND(v4l2src,
					CAMD_MSG_ID_N_PRE_STILL,
					CAMD_MSG_ARG_STATE_OFF)) {
			goto error_ipc_failed;
		}
	}

	GST_INFO_OBJECT(v4l2src, "starting still image capture");

	PROF_BEGIN(still_exposure_and_readout);

	gint type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	gint fd = v4l2src->v4l2object->still_fd;

	if (!gst_v4l2src_capture_stop(v4l2src))
		goto stop_preview_failed;
	if (!gst_v4l2src_set_still_capture(v4l2src))
		goto s_still_capture_failed;
	if (!queue_still_buffer(v4l2src))
		goto queue_buffer_failed;
	PROF_SECT_DECL_BEGIN(wait_for_camd_still_cfg);
	if (CAMD_NOTIFICATION_SEND(v4l2src,
				CAMD_MSG_ID_N_STILL_CAPTURE,
				CAMD_MSG_ARG_STATE_ON)) {
		goto error_ipc_failed;
	}
	PROF_SECT_END(wait_for_camd_still_cfg);
	if (ioctl(fd, VIDIOC_STREAMON, &type) < 0)
		goto streamon_failed;
	gst_v4l2src_send_appmsg(v4l2src, CAPTURE_START);
	ret = GST_FLOW_OK;
	goto gst_v4l2src_still_capture_start_exit;
streamon_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, OPEN_READ,
				(_("Error during streamon of still capture, errno=%d"),
				 errno), GST_ERROR_SYSTEM);
		goto gst_v4l2src_still_capture_start_exit;
	}
queue_buffer_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("Error queueing buffer into device '%s'."),
				 v4l2src->v4l2object->videodev), GST_ERROR_SYSTEM);
		goto gst_v4l2src_still_capture_start_exit;
	}
s_still_capture_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("set still capture failed on device '%s'"),
				 v4l2src->v4l2object->videodev),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_still_capture_start_exit;
	}
stop_preview_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("Error stopping prevew from device '%s'."),
				 v4l2src->v4l2object->videodev),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_still_capture_start_exit;
	}
error_ipc_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, OPEN_READ,
				(_("Error communicating with camd (is it alive?)")),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_still_capture_start_exit;
	}
gst_v4l2src_still_capture_start_exit:
	return ret;
}


static void handle_still_capture_notification(GstV4l2Src *v4l2src)
{
	GST_DEBUG("capture notification received");
	PROF_END(start_still_capture_lag);

	g_assert(FCM_PREVIEW_1ST_FRAME != v4l2src->frame_create_mode);
	v4l2src->still_buffer_ready = FALSE;

	if (FCM_PREVIEW == v4l2src->frame_create_mode) {
#ifdef STORE_PRE_STILL_FRAMES
		fstore_session_begin(v4l2src->preview_fh);
#endif			
		if (v4l2src->flash_ctrl != FLASH_OFF) {
			PROF_BEGIN(torch_converge);
			if (CAMD_NOTIFICATION_REQ_DATA(v4l2src, CAMD_MSG_ID_N_PRE_STILL,
						CAMD_MSG_ARG_STATE_ON,
						&v4l2src->flash_required,
						sizeof(v4l2src->flash_required))) {
				GST_ELEMENT_ERROR(v4l2src, RESOURCE, SYNC,
						(_("Error sending camd pre-still event")),
						GST_ERROR_SYSTEM);
			}
		}
	}

	/* sample and store the flash ctrl for the current
	 * still capture sequence
	 */
	v4l2src->current_flash_ctrl = v4l2src->flash_ctrl;
	if (v4l2src->current_flash_ctrl == FLASH_OFF ||
			(v4l2src->current_flash_ctrl == FLASH_AUTO &&
			 !v4l2src->flash_required)) {
		v4l2src->frame_create_mode = FCM_STILL_CAPTURE;
		if (gst_v4l2src_still_capture_start(v4l2src) != GST_FLOW_OK)
			GST_WARNING_OBJECT(v4l2src, "cannot start still capture");
	}
	else {
		v4l2src->frame_create_mode = FCM_PREVIEW_W_TORCH;
#ifdef SHOW_TORCH_CONVERGE_SEQ
		GST_DEBUG("current_flash_ctrl=%d, flash_required=%d",
				v4l2src->current_flash_ctrl,
				v4l2src->flash_required);
#endif
	}
}

static GstFlowReturn get_torched_preview_buf(GstV4l2Src *v4l2src, GstBuffer **buf)
{
	int converged;
	GstFlowReturn ret = GST_FLOW_OK;

	if (CAMD_NOTIFICATION_REQ_DATA(v4l2src,
				CAMD_MSG_ID_Q_IS_PREFLASH_OK,
				CAMD_MSG_ARG_NONE,
				&converged,
				sizeof(converged))) {
		return GST_FLOW_ERROR;
	}
#ifdef SHOW_TORCH_CONVERGE_SEQ
	GST_DEBUG("camd flash convergece state=%d", converged);
#endif					
	if (converged) {
		v4l2src->frame_create_mode = FCM_STILL_CAPTURE;
		if (gst_v4l2src_still_capture_start(v4l2src) != GST_FLOW_OK) {
			GST_WARNING_OBJECT(v4l2src, "cannot start still capture");
			return GST_FLOW_ERROR;
		}
		PROF_END(torch_converge);
	}
	else
		ret = gst_v4l2src_grab_video_frame(v4l2src, buf);

	return ret;
}


/* Values for bit 0 indicating whether the flash fired.
 * 0 = Flash did not fire
 * 1 = Flash fired
 *
 * Values for bits 1 and 2 indicating the status of returned light.
 * 00 = No strobe return detection function
 * 01 = reserved
 * 10 = Strobe return light not detected
 * 11 = Strobe return light detected
 *
 * Values for bits 3 and 4 indicating the camera's flash mode.
 * 00 = unknown
 * 01 = Compulsory flash firing
 * 10 = Compulsory flash suppression
 * 11 = Auto mode
 *
 * Values for bit 5 indicating the presence of a flash function.
 * 0 = Flash function present
 * 1 = No flash function
 *
 * Values for bit 6 indicating the camera's red-eye mode.
 * 0 = No red-eye reduction mode or unknown
 * 1 = Red-eye reduction supported
 */

enum exif_flash_fields {
	FLASH_FIRED = 0,
	FLASH_RETURNED_LIGHT = 1,
	FLASH_MODE = 3,
	FLASH_PRESENT = 5,
	FLASH_REDEYE_SUPPORT = 6,
};

#define FLASH_SET_FIELD(f, v) (v << f)

static void gst_v4l2src_update_still_exif(GstV4l2Src *v4l2src)
{
	ImageBuffer *data = (ImageBuffer *)GST_BUFFER_DATA(v4l2src->still_capture_buffer);
	photoData *photo = &data->exif;

	unsigned int	exifv = 0x30323230;
	gboolean flash_on = v4l2src->current_flash_ctrl == FLASH_ON ||
		(v4l2src->current_flash_ctrl == FLASH_AUTO &&
		 v4l2src->flash_required);
	int flash_mode;
	switch (v4l2src->current_flash_ctrl) {
		case FLASH_ON: flash_mode = 1; break;
		case FLASH_OFF: flash_mode = 2; break;
		case FLASH_AUTO: flash_mode = 3; break;
		default:
				 g_assert(0);
	}

	photo->Flash = FLASH_SET_FIELD(FLASH_FIRED, flash_on) |
		FLASH_SET_FIELD(FLASH_MODE, flash_mode) |
		FLASH_SET_FIELD(FLASH_PRESENT, 0);

	photo->XResolution[0] = 72;
	photo->XResolution[1] = 1;
	photo->YResolution[0] = 72;
	photo->YResolution[1] = 1;
	photo->ResolutionUnit = 2;
	photo->PixelXDimension = 1520;
	photo->PixelYDimension = 2048;
	photo->ExposureTime[0] = v4l2src->psd.exposure_time[0];
	photo->ExposureTime[1] = v4l2src->psd.exposure_time[1];
	photo->ApertureValue[0] = 10;
	photo->ApertureValue[1] = 56;
	photo->FocalLength[0] = 325;
	photo->FocalLength[1] = 100;
	photo->YCbCrPositioning = 1; // centered
	photo->FNumber[0] = 24;	
	photo->FNumber[1] = 10;
	photo->ExposureProgram = 2;
	photo->ComponentsConfiguration[0] = 0;
	photo->ComponentsConfiguration[1] = 3;
	photo->ComponentsConfiguration[2] = 2;
	photo->ComponentsConfiguration[3] = 1;
	photo->ExposureMode = 0;
	photo->WhiteBalance = 0;
	photo->valid = IMG_PIXX_MASK
		| IMG_PIXY_MASK
		| IMG_VER_MASK
		| IMG_APPVAL_MASK
		| IMG_FOCLN_MASK
		| IMG_EXPTIM_MASK
		| IMG_YCBCR_MASK
		| IMG_FNUMBR_MASK
		| IMG_EXP_MASK
		| IMG_COMP_MASK
		| IMG_FLASH_MASK
		| IMG_EXPPROG_MASK
		| IMG_WHTBAL_MASK
		| IMG_XRES_MASK
		| IMG_YRES_MASK
		| IMG_RESU_MASK;
	memcpy(photo->ExifVersion,&exifv, 4);

	GST_DEBUG_OBJECT(v4l2src, "Captured Driver EXIF Data");
	GST_DEBUG_OBJECT(v4l2src, "...photo->XResolution = %d / %d",
			photo->XResolution[0], photo->XResolution[1]);
	GST_DEBUG_OBJECT(v4l2src, "...photo->YResolution = %d / %d",
			photo->YResolution[0], photo->YResolution[1]);
	GST_DEBUG_OBJECT(v4l2src, "...photo->ResolutionUnit = %d", photo->ResolutionUnit);
	GST_DEBUG_OBJECT(v4l2src, "...photo->PixelXDimension = %d", photo->PixelXDimension);
	GST_DEBUG_OBJECT(v4l2src, "...photo->PixelYDimension = %d", photo->PixelYDimension);
	GST_DEBUG_OBJECT(v4l2src, "...photo->ExposureTime = %d / %d ( 1 / %f s )",
			photo->ExposureTime[0],
			photo->ExposureTime[1],
			(float)photo->ExposureTime[1] / (float)photo->ExposureTime[0] );
	GST_DEBUG_OBJECT(v4l2src, "...photo->ApertureValue = %d / %d (F %f)",
			photo->ApertureValue[0],
			photo->ApertureValue[1],
			(float) photo->ApertureValue[1] / (float) photo->ApertureValue[0]);
	GST_DEBUG_OBJECT(v4l2src, "...photo->FocalLength = %d / %d (%f)",
			photo->FocalLength[0],
			photo->FocalLength[1],
			(float)photo->FocalLength[0]/(float)photo->FocalLength[1]);
	GST_DEBUG_OBJECT(v4l2src, "...photo->YCbCrPositioning = %d",
			photo->YCbCrPositioning);
	GST_DEBUG_OBJECT(v4l2src, "...photo->FNumber = %d / %d (F %f)",
			photo->FNumber[0],
			photo->FNumber[1],
			(float)photo->FNumber[0] / (float)photo->FNumber[1]);

	GST_DEBUG_OBJECT(v4l2src, "...photo->ExposureProgram = %d", photo->ExposureProgram);
	GST_DEBUG_OBJECT(v4l2src, "...photo->ComponentsConfiguration[0] = %d",
			photo->ComponentsConfiguration[0]);
	GST_DEBUG_OBJECT(v4l2src, "...photo->ComponentsConfiguration[1] = %d",
			photo->ComponentsConfiguration[1]);
	GST_DEBUG_OBJECT(v4l2src, "...photo->ComponentsConfiguration[2] = %d",
			photo->ComponentsConfiguration[2]);
	GST_DEBUG_OBJECT(v4l2src, "...photo->ComponentsConfiguration[3] = %d",
			photo->ComponentsConfiguration[3]);
	GST_DEBUG_OBJECT(v4l2src, "...photo->ExposureMode = %d", photo->ExposureMode);
	GST_DEBUG_OBJECT(v4l2src, "...photo->WhiteBalance = %d", photo->WhiteBalance);
	GST_DEBUG_OBJECT(v4l2src, "...photo->valid = 0x%x", photo->valid);
}

/* returns TRUE if resumed externally, FALSE otherwise */
static gboolean pause_post_still_capture(GstV4l2Src *v4l2src)
{
	gboolean ret = FALSE;
	GST_OBJECT_LOCK(v4l2src);
	GstClockTime diff = gst_clock_get_time(GST_ELEMENT_CLOCK(v4l2src)) -
		v4l2src->still_capture_start_ts;
	GST_OBJECT_UNLOCK(v4l2src);
	gint net_still_capture_ms = GST_TIME_AS_MSECONDS(diff);
	gint sleep_ms = v4l2src->still_duration - net_still_capture_ms;

	if (sleep_ms > 0) {
		g_mutex_lock(v4l2src->post_still_hold_frame_lock);
		GST_DEBUG("sleeping additional %dms to achive a total pause of %dms",
				sleep_ms, v4l2src->still_duration);
		v4l2src->pausing_after_still = TRUE;
		GTimeVal abs_time;
		g_get_current_time(&abs_time);
		g_time_val_add(&abs_time, sleep_ms * 1000);
		ret = g_cond_timed_wait(v4l2src->still_hold_frame_cond,
				v4l2src->post_still_hold_frame_lock, &abs_time);
		GST_DEBUG(ret ? "resumed externally" : "completed requested wait");
		v4l2src->pausing_after_still = FALSE;
		g_mutex_unlock(v4l2src->post_still_hold_frame_lock);
	}
	else {
		GST_WARNING("requested pause time (%dms) is below net still capture "
				"time (%dms), sleep skipped",
				v4l2src->still_duration, net_still_capture_ms);
	}
	return ret;
}

static GstFlowReturn gst_v4l2src_still_capture_complete(GstV4l2Src *v4l2src)
{
	GstFlowReturn ret = GST_FLOW_ERROR;
	gint fd = v4l2src->v4l2object->still_fd;

	GST_DEBUG("about to grab still frame from fd=%d", fd);

	if (GST_FLOW_OK != gst_v4l2src_grab_still_frame(v4l2src))
		goto grab_failed;

	PROF_END(still_exposure_and_readout);

	if (CAMD_NOTIFICATION_REQ_DATA(v4l2src,
				CAMD_MSG_ID_N_STILL_CAPTURE,
				CAMD_MSG_ARG_STATE_OFF,
				&v4l2src->psd, sizeof(v4l2src->psd))) {
		goto error_ipc_failed;
	}

	gint type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(fd, VIDIOC_STREAMOFF, &type) < 0)
		goto streamoff_failed;

	gst_v4l2src_update_still_exif(v4l2src);
	GST_DEBUG_OBJECT(v4l2src, "issuing gst_pad_push on still pad, refcount=%d",
			GST_MINI_OBJECT_REFCOUNT_VALUE(v4l2src->still_capture_buffer));
	PROF_SECT_DECL_BEGIN(pad_push_to_vpp);
#ifdef STORE_STILL_FRAMES
	fstore_save_buf(v4l2src->still_fh, v4l2src->still_capture_buffer, 1);
#endif
	GstBuffer *tmp = GST_BUFFER(v4l2src->still_capture_buffer);
	/* we loose the ref after the push, might as well invalidate it */
	v4l2src->still_capture_buffer = NULL;
	ret = gst_pad_push(v4l2src->src_still_pad, tmp);
	GST_DEBUG("after pad_push, still capture buffer refount=%d ",
			GST_MINI_OBJECT_REFCOUNT_VALUE(tmp));

	if (ret != GST_FLOW_OK)
		goto gst_pad_push_failed;
	PROF_SECT_END(pad_push_to_vpp);

	if (v4l2src->still_duration && pause_post_still_capture(v4l2src)) {
		/* nothing useful to do here, really */
	}
	PROF_SECT_DECL_BEGIN(back_to_preview_after_still_complete);
	GST_V4L2_SET_INACTIVE(v4l2src->v4l2object);
	if (!gst_v4l2src_set_capture(v4l2src,
				     v4l2src->pixelformat,
				     v4l2src->width,
				     v4l2src->height,
				     v4l2src->fps_n, v4l2src->fps_d))
		goto gst_v4l2src_still_capture_complete_exit;
	GST_V4L2_SET_ACTIVE(v4l2src->v4l2object);

	if (!gst_v4l2src_capture_start(v4l2src))
		goto gst_v4l2src_still_capture_complete_exit;
	PROF_SECT_END(back_to_preview_after_still_complete);
	PROF_END(still_start_until_ready);
	ret = GST_FLOW_OK;
	goto gst_v4l2src_still_capture_complete_exit;

gst_pad_push_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("gst_pad_push() returned %d"), ret), GST_ERROR_SYSTEM);
		goto gst_v4l2src_still_capture_complete_exit;
	}
streamoff_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("Error stopping streaming still capture "
				   "from device '%s'."),
				 v4l2src->v4l2object->videodev),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_still_capture_complete_exit;
	}
error_ipc_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, OPEN_READ,
				(_("Error communicating with camd (is it alive?)")),
				GST_ERROR_SYSTEM);
		goto gst_v4l2src_still_capture_complete_exit;
	}
grab_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_("grab still frame from device '%s' failed."),
				 v4l2src->v4l2object->videodev),
				GST_ERROR_SYSTEM);
		if (ioctl(fd, VIDIOC_STREAMOFF, &type) < 0)
			goto streamoff_failed;
		goto gst_v4l2src_still_capture_complete_exit;
	}
gst_v4l2src_still_capture_complete_exit:
	return ret;
}

static void update_flash_mode(GstV4l2Src *v4l2src)
{
	camd_msg_arg_t mode;

	switch (v4l2src->flash_ctrl) {
		case FLASH_OFF:
			mode = CAMD_MSG_ARG_STATE_OFF;
			break;
		case FLASH_ON:
			mode = CAMD_MSG_ARG_STATE_ON;
			break;
		case FLASH_AUTO:
			mode = CAMD_MSG_ARG_STATE_AUTO;
			break;
		default:
			g_assert(0);
			break;
	}
	CAMD_NOTIFICATION_SEND(v4l2src, CAMD_MSG_ID_N_FLASH, mode);
}

static GstFlowReturn gst_v4l2src_create(GstPushSrc *src, GstBuffer **buf)
{
	GstV4l2Src *v4l2src = GST_V4L2SRC(src);
	GstFlowReturn ret = GST_FLOW_OK;
	g_assert(v4l2src->init_complete);
	*buf = NULL;

#ifdef ENABLE_PROFILING
	{
		static int preview_events_count;
		if (!(preview_events_count++ % 300))
			profile_print_stats(0);
	}
#endif
	PROF_BEGIN(wait_for_preview_frame);

	/* if new preview frame is ready, return it */
	while (!*buf) {
		short revents;
		while (1) {
			if (v4l2src->frame_create_mode == FCM_STILL_CAPTURE) {
				setup_events_still(v4l2src);
			}
			else {
				if (queue_available_preview_buffers(v4l2src) != GST_FLOW_OK)
					goto queue_buffers_error;
				setup_events_preview(v4l2src);
			}

			switch (get_events(v4l2src)) {
				case GE_EVENTS_RECEIVED:
					break;
				case GE_TIMEOUT:
					continue;
				case GE_FATAL_ERROR:
					goto events_error;
				default:
					g_assert(0);
			}
			if (EVENTS(PE_NOTIFY) & POLLIN) {
				gint msg;
				CAMSRC_R_NOTIFICATION(&msg);
				switch (msg) {
					case CAMSRC_N_STILL_CAPTURE:
#ifdef STORE_STILL_FRAMES
						fstore_session_begin(v4l2src->still_fh);
#endif			
						handle_still_capture_notification(v4l2src);
						break;
					case CAMSRC_N_BUFFER_READY:
#ifdef SHOW_BUFFER_FINILIZE_NOTIFICATIONS
						GST_DEBUG("buffer finalize notification "
								"received");
#endif					
						continue;
					case CAMSRC_N_PAUSED:
						GST_DEBUG("plugin paused");
						return GST_FLOW_WRONG_STATE;
					case CAMSRC_N_FLASH_MODE:
						update_flash_mode(v4l2src);
						continue;
					default:
						g_assert(0);
				}
				break;
			}
			if (EVENTS(PE_V4L2) & POLLIN)
				break;
			if ((revents = EVENTS(PE_IPC))) {
				if (revents & POLLHUP)
					goto camd_hup_detected;

				/* unhandled traffic on camd ipc channel is unexpected */
				g_assert(!(revents & POLLIN));
			}
			break;
		}

		/* check if capture command received/in progress */
		switch (v4l2src->frame_create_mode) {
			case FCM_PREVIEW_1ST_FRAME:
				v4l2src->frame_create_mode = FCM_PREVIEW;
				still_capture_event_notify(v4l2src);

				/* fallthrough */
			case FCM_PREVIEW:
				ret = gst_v4l2src_grab_video_frame(v4l2src, buf);
				PROF_END(wait_for_preview_frame);
				if (v4l2src->num_discard_video_frames) {
					v4l2src->num_discard_video_frames--;
					gst_buffer_unref(*buf);
					*buf = NULL;
					continue;
				}
				break;
			case FCM_PREVIEW_W_TORCH:
				ret = get_torched_preview_buf(v4l2src, buf);
				break;
			case FCM_STILL_CAPTURE:
				GST_DEBUG("completing still capture");
				ret = gst_v4l2src_still_capture_complete(v4l2src);
				if (ret != GST_FLOW_OK)
					goto gst_v4l2src_create_exit;
#ifdef STORE_STILL_FRAMES
				fstore_session_end(v4l2src->still_fh);
#endif			
				v4l2src->frame_create_mode = FCM_PREVIEW_1ST_FRAME;
#ifdef STORE_PRE_STILL_FRAMES
				fstore_session_end(v4l2src->preview_fh);
#endif
				break;
		}
	}
	/* frame should be ready */
	g_assert(*buf);

#ifdef SHOW_FRAMES_QDQ
	GST_LOG("pushing buffer addr=%p data=%p timestamp=%lld", 
			*buf, GST_BUFFER_DATA(*buf), GST_BUFFER_TIMESTAMP(*buf));
#endif
	goto gst_v4l2src_create_exit;

queue_buffers_error:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS, (_("buffers queue failed")),
				GST_ERROR_SYSTEM);
		ret = GST_FLOW_ERROR;
		goto gst_v4l2src_create_exit;
	}
events_error:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS, (_("events error")),
				GST_ERROR_SYSTEM);
		ret = GST_FLOW_ERROR;
		goto gst_v4l2src_create_exit;
	}
camd_hup_detected:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				(_("camd-ipc channel closed unexpectedly")),
				GST_ERROR_SYSTEM);
		ret = GST_FLOW_ERROR;
		goto gst_v4l2src_create_exit;
	}   

gst_v4l2src_create_exit:
#ifdef STORE_PRE_STILL_FRAMES
	fstore_save_buf(v4l2src->preview_fh, *buf, 0);
#endif

	if (*buf) {
		GST_TSDEBUG("TSDEBUG: CAM GST Out: %lld (%lld)", 
				GST_BUFFER_TIMESTAMP(*buf), GST_BUFFER_DURATION(*buf)); 
#ifdef YUVS_ENABLE
		int val;
		g_assert(!sem_getvalue(&v4l2src->yuvs.sem, &val));
		if (val < YUVS_MAX_QUQUE_SIZE) {
			gst_buffer_ref(*buf);
			g_mutex_lock(v4l2src->yuvs.lock);
			g_queue_push_head(v4l2src->yuvs.queue, *buf);
			g_mutex_unlock(v4l2src->yuvs.lock);
			g_assert(!sem_post(&v4l2src->yuvs.sem));
		}
#endif
	}

	return ret;
}
